# ------------------------------------------------------------------
# Licensed under the MIT License. See LICENSE in the project root.
# ------------------------------------------------------------------

"""
    Rotate(rot)

Rotate geometry or mesh with rotation `rot`
from Rotations.jl.

## Examples

```julia
Rotate(one(RotXYZ{Float64}))  # Generate identity rotation
Rotate(AngleAxis(0.2, 1.0, 0.0, 0.0))  # Rotate 0.2 radians around X-axis
Rotate(rand(QuatRotation{Float64}))  # Generate random rotation
```
"""
struct Rotate{R<:Rotation} <: CoordinateTransform
  rot::R
end

"""
    Rotate(u, v)

Rotation mapping the axis directed by `u` to the axis directed by `v`. 
More precisely, it maps the plane passing through the origin with normal 
vector `u` to the plane passing through the origin with normal vector `v`.

## Examples

```julia
Rotate(Vec(1, 0, 0), Vec(1, 1, 1))
```
"""
Rotate(u::Vec, v::Vec) = Rotate(rotation_between(u, v))

Rotate(u::Tuple, v::Tuple) = Rotate(Vec(u), Vec(v))

isrevertible(::Type{<:Rotate}) = true

isinvertible(::Type{<:Rotate}) = true

Base.inv(t::Rotate) = Rotate(inv(t.rot))

applycoord(t::Rotate, v::Vec) = t.rot * v

# --------------
# SPECIAL CASES
# --------------

applycoord(t::Rotate, b::Box{2}) = applycoord(t, convert(Quadrangle, b))

applycoord(t::Rotate, b::Box{3}) = applycoord(t, convert(Hexahedron, b))

applycoord(t::Rotate, g::Grid) = applycoord(t, convert(SimpleMesh, g))
